"""The extraction module contains structures for handling signal extraction
from imaging datasets, and related problems (identifying ROIs,
calculating traces, etc...).

Currently implemented are routines for using the Suite2p package
(lab3.extraction.s2p). In the future, this could be expanded to handle other
pipelines (e.g. CaImAn, manual segmentation...)"""
