"""This submodule contains analysis routines for the analysis of spatial
tuning."""

from lab3.analysis.spatial.spatial_tuning import (
    SimpleSpatialTuning,
    CircularSimpleSpatialTuning
)

from lab3.analysis.spatial.place_fields import *
